# txrx.tcl --
#
#       The procedures in this script act as a wrapper to native routines
#       of txrx and provide a convenient interface to RPC v2 for
#       the user of the package.
#
# SCCS: @(#) txrx.tcl 1.1 96/08/13 15:27:00
#
# See the file "license.terms" for information on usage and redistribution
# of this file, and for a DISCLAIMER OF ALL WARRANTIES.

package provide txrx 1.0

#
# Global variable - the array txrx.
# txrx has indices of the format "protocolName,keyword" where
# protocolName is the name of the protocol as specified in
# a call to txrxCreate and keyword is one of the strings below:
#
# xid - provides the current sequence number for an RPC call.
# auth_credentials - credentials for the current RPC call.
# auth_verifier - verifier for the current RPC call.
# wait - variable to wait for (postfixed with channel name)
# bind - binding (Tcl command) for protocol procedure
# timeout,count - number of times to wait for an answer from the server
# timeout,delay - time (in miliseconds) to sleep while waiting for an
#                 answer from the server.
# channel,host,program,version - cache for open channels to other servers
#                                used by txrxCall. On txrxDelete the
#                                channels are closed.
# registered,program,version - used to store service port number so that
#                          deregistration of the service is possible
#                          when closing the protocol.

# For illustrative purposes only:
set txrx(xid) 0
set txrx(auth_credentials) [list AUTH_NONE {}]
set txrx(auth_verifier) [list AUTH_NONE {}]
set txrx(reply_verifier) {}

# txrxCreate --
#   Create an RPC protocol description.
#
# Parameters:
#   protocol name - the name desired to represent the protocol, optional
#   source - the text source of the protocol in a format acceptable
#            to txrx parse
#
# Result:
#   name of the protocol as returned by txrx parse

proc txrxCreate {args} {
    if {[llength $args] != 1 && [llength $args] != 2} {
	return -code error "wrong # args: txrxCreate ?protocolName? string"
    }
    if {[llength $args] == 1} {
	set name  [txrx parse [lindex $args 0]]
    } else {
        set name [txrx parse [lindex $args 1] [lindex $args 0]]
    }
    global txrx
    set txrx($name,xid) 0
    set txrx($name,auth_credentials) [list AUTH_NONE {}]
    set txrx($name,auth_verifier) [list AUTH_NONE {}]
    set txrx($name,reply_verifier) {}
    set txrx($name,timeout) "3 4000"
    return $name
}

# txrxTimeout --
#   Setup timeout for a given protocol.
#
# Parameters:
#   protocol
#   count - number of retries (retransmits for a packet oriented protocol)
#   delay - delay between retries
#
# Result:
#   None.

proc txrxTimeout {protocol count delay} {
    global txrx
    if [catch "set txrx($protocol,xid)"] {
	return -code error "no such protocol $protocol" 
    }
    # Make sure these are numbers.
    incr count 0
    incr delay 0
    set txrx($protocol,timeout) [list $count $delay]
}
    
# txrxDelete --
#   Delete an RPC protocol description.
#
# Parameters:
#   protocol name - the name desired to represent the protocol, optional
#
# Result:
#   None.

proc txrxDelete {args} {
    global txrx
    foreach arg $args {
        # Get the set of open channels related to the protocol.
        set chanNames [array names txrx $arg,channel,*]
        foreach chanName $chanNames {
            set chan $txrx($chanName)
            # Some channels may have been closed already, although
            # user should have no knowledge of them.
            catch "close $chan"
            
        }
        set serviceNames [array names txrx $arg,registered,*]
        foreach serviceName $serviceNames {
              # This may fail if the protocol name contains comms.
              # For the moment leave it as is.
              set l [split $serviceName ","]
              set sz [llength $l]
              set program [lindex $l [expr $sz - 2]]
              set version [lindex $l [expr $sz - 1]]
              catch "txrxDeregister $arg $program $version" 
        }
        # Delete all variables in array txrx associated to the protocol.
        set argNames [array names txrx $arg,*]
        foreach n $argNames {
	    unset txrx($n)
	}
        # Finally, delete the protocol description itself.
        catch "txrx delete $arg"
    }
}

# txrxList --
#   List all RPC protocol descriptions.
#
# Parameters:
#
# Result:
#   names of protocols as returned by txrx list

proc txrxList {} {
    if [catch "txrx list" lst] {
	return -code error "txrx extension is not loaded"
    }
    return $lst
}

# txrxBind --
#   Setup a Tcl command to call when servicing a given RPC procedure.
#   Inquire about existing bindings.
#
# Parameters:
#   protocol - name of protocol description
#   procedure - protocol procedure name
#   cmd - Tcl command string
# Result:
#   the name of the procedure is "canonical" -- proc ver prog -- format.

proc txrxBind {args} {
    global txrx
    if {[llength $args] != 1 && [llength $args] != 3} {
        return -code error \
		"wrong # args: txrxBind protocol procedure commandString"
    }
    if {[llength $args] == 3} {
	return [eval txrxBind1 $args]
    }
    set res {}
    set protocol [lindex $args 0]
    foreach binding [array names txrx $protocol,bind,*] {
       lappend res $txrx($binding)
    }
    return $res
}

# txrxBind1 --
#   Setup a Tcl command to call when servicing a given RPC procedure.
#
# Parameters:
#   protocol - name of protocol description
#   procedure - protocol procedure name
#   cmd - Tcl command string
# Result:
#   the name of the procedure is "canonical" -- proc ver prog -- format.

proc txrxBind1 {protocol procedure cmd} {
    global txrx
    # Get the canonical name for the procedure.
    set procedureName [lindex [$protocol info procedure $procedure] 0]
    # Store the binding as an entry in the array.
    set proc [lindex $procedureName 0]
    set vers [lindex $procedureName 1]
    set prog [lindex $procedureName 2]
    set txrx($protocol,bind,$proc,$vers,$prog) [list $procedureName $cmd]
    return $procedureName
}

# txrxService --
#   Service an RPC call.
#
# Parameters:
#   protocol - name of the RPC protocol.
#   chan     - name of channel where the call comes from
# 
# Result:
#   Normally an empty string. In case an error occured during service
#   errorInfo will contain human readable information - it may be
#   useful to log this information. Same about errorCode - it can be
#   sampled by the application and interpreted.
#   The behavior is intentional!
#   txrxService is called usually during background processing, namely
#   during a fileevent callback. This means the only means of reporting
#   errors is via bgerror. Most programs, when first developed, do not 
#   provide a bgerror procedure and simple error conditions
#   like a failed authorization or a channel close would cause bgerror
#   to be invoked spuriously. Thus the decision was to catch all possible
#   errors that may occur during processing and return information
#   about errors in errorInfo and errorCode. The server developer should
#   check if [lindex $errorCode 0] is equal to "TXRX" - in that case
#   bgerror or other error procedures may be invoked explicitly.

proc txrxService {protocol chan} {
    global txrx
    # Read RPC header.
    if [catch "$protocol buffer init read $chan" res] {
        # There is no data available to read, the
        # client has closed the channel.
        $protocol buffer reset
        return -errorinfo $res \
		-errorcode {TXRX READ_DATA "cannot read channel"}
    }
    if [catch "$protocol read __rpc" header] {
        # Send an error reply stating that a system error occured.	
	txrxSendErrorReply $protocol $chan \
		{ 0 {REPLY {MSG_ACCEPTED {"AUTH_NONE {}" {SYSTEM_ERR {}}}}}}
        return -errorinfo $res \
		-errorcode {TXRX READ_HEADER "cannot read call header"}
    }
    set xid [lindex $header 0]
    # Find binding between procedure and Tcl command, get information
    # about the procedure.
    if [catch "txrxInterpretCallHeader $protocol [list $header]" info] {
        # An error occured while processing the header.
        txrxSendErrorReply $protocol $chan [list $xid $info]
        return -errorcode {TXRX INTERPRET_HEADER "cannot interpret header"}
    }
    # Check if a binding exists for the procedure.
    set procName [lindex $info 0]
    set proc [lindex $procName 0]
    set vers [lindex $procName 1]
    set prog [lindex $procName 2]
    if [catch "set txrx($protocol,bind,$proc,$vers,$prog)" binding] {
        # Send an error reply stating that no procedure is available.
        txrxSendErrorReply $protocol $chan [list $xid \
                {REPLY {MSG_ACCEPTED {"AUTH_NONE {}" {PROC_UNAVAIL {}}}}}]
        return -errorcode {TXRX PROC_UNAVAIL "no procedure available"}
    }
    # Prepare arguments.
    if [catch "txrxReadCallArguments $protocol [list $info]" res] {
        # An error occured while reading the call arguments.
        txrxSendErrorReply $protocol $chan [list $xid $res]
        return -errorcode {TXRX GARBAGE_ARGS "garbage arguments"}
    }
    # Call the Tcl procedure.
    set cmd [lindex $binding 1]

    if [catch "$cmd $res" callResult] {
        # An unknown error has occured during execution of the procedure.
        txrxSendErrorReply $protocol $chan [list $xid \
                {REPLY {MSG_ACCEPTED {"AUTH_NONE {}" {SYSTEM_ERR {}}}}}]
        return -errorcode {TXRX SYSTEM_ERR "error while executing procedure"}
    }
    # Prepare reply header.
    catch "$protocol buffer reset"
    catch "$protocol buffer init write $chan"
    if [catch "$protocol write __rpc [list [list $xid \
	{REPLY {MSG_ACCEPTED {"AUTH_NONE {}" {SUCCESS {}}}}}]]" res] {

	txrxSendErrorReply $protocol $chan [list $xid \
		{REPLY {MSG_ACCEPTED {"AUTH_NONE {}" {SYSTEM_ERR {}}}}}]
        return -errorcode {TXRX SYSTEM_ERR "error while sending reply header"}
    }

    # Convert result to network format.
    if [catch "$protocol write \
	[lindex [lindex $info 1] 0] [list $callResult]" err] {
        # Result cannot be converted to binary.
        txrxSendErrorReply $protocol $chan [list $xid \
                [list REPLY [list MSG_ACCEPTED [list "AUTH_NONE {}" \
                                                    [list SYSTEM_ERR {}]]]]]
        return -errorcode {TXRX SYSTEM_ERR "error while sending call results"}
    }
    # Send reply to caller.
    catch "$protocol buffer flush"
    catch "flush $chan"
    catch "$protocol buffer reset"
    return ""
}

#  
#
# txrxCallDirect --
#   Execute an RPC call to a server.
# 
# Arguments:
#   protocol   - txrx protocol description generated by a "txrx parse" command
#   channel    - channel where communication takes place
#   procedure  - name or canonical representation for an RPC procedure
#   args       - whatever arguments need to be sent to the server
#
# Result:
#   the result is the result of the remote call or an error message
#   in case of error

proc txrxCallDirect {protocol channel procedure args} {
    global txrx
    $protocol buffer init write $channel
    # Prepare call header.
    set procedureInfo [$protocol info procedure $procedure]
    set callHeader [txrxPrepareHeader $protocol [lindex $procedureInfo 0]]
    $protocol write __rpc $callHeader
    # Prepare arguments.
    set argTypes [lrange [lindex $procedureInfo 1] 1 end]
    set i 0
    foreach argType $argTypes {
        set matchingArg [lindex $args $i]
        $protocol write $argType $matchingArg
	if {$argType == 1} {
	    # VOID format arguments are not supposed to have a matching
	    # actual argument. Use the same actual argument for the
            # next formal argument.
        } else {
	    # Non-VOID argument. Must match the type provided.
            incr i
        }
    }
    $protocol buffer flush
    # Perform RPC call (basically flush the output buffer).
    flush $channel
    $protocol buffer reset
    # Need to set up a timeout using fileevent and after.
    set repeatCountMax [lindex $txrx($protocol,timeout) 0]
    set repeatCount 0
    set snoozeTime  [lindex $txrx($protocol,timeout) 1]
    set saveEvent [fileevent $channel readable]

    while {$repeatCount < $repeatCountMax && $repeatCount >= 0} {
  	fileevent $channel readable "set txrx($protocol,wait,$channel) 1"
        set idTimer [after $snoozeTime "set txrx($protocol,wait,$channel) 0"]

        vwait txrx($protocol,wait,$channel)

        # Timeout or data available.
        if {$txrx($protocol,wait,$channel) == 0} {
	    incr repeatCount
 	} else {
            # Data is available, cancel timer event.
            after cancel $idTimer
	    set repeatCount -1
        }
    }
    # Regardless of the outcome, we need to restore the initial
    # file event on the channel.
    fileevent $channel readable $saveEvent
    if {$repeatCount != -1} {
        # Repeated timeout, server seems to be gone south.
        $protocol buffer reset
        return -code error -errorinfo "timeout" "timeout after retry"
    }
     
    if [catch "$protocol buffer init read $channel" err] {
	$protocol buffer reset
        return -code error $err
    }
    # Receive header.
    set replyHeader [$protocol read __rpc]
    if [catch "txrxInterpretReplyHeader $protocol [list $replyHeader]" err] {
        $protocol buffer flush
        $protocol buffer reset
        return -code error $err 
    }
    # Receive result.
    set resultType [lindex [lindex $procedureInfo 1] 0]
    if [catch "$protocol read $resultType" result] {
        $protocol buffer flush
        $protocol buffer reset
        return -code error -errorinfo $result "error receiving result"
    }
    # Reset buffer
    $protocol buffer flush
    $protocol buffer reset
    return $result
}

# 
# txrxPrepareHeader --
#   Setup an RPC call header suitable for conversion to wire format.
#
# Arguments:
#   name - procedure name in canonical "proc version program" format
#   
# Result:
#   list with call header

proc txrxPrepareHeader {protocol name} {
    global txrx
    set xid $txrx($protocol,xid)
    set procNumber [lindex $name 0]
    set verNumber [lindex $name 1]
    set progNumber [lindex $name 2]
    return [list $xid [list CALL [list 2 $progNumber $verNumber $procNumber \
                                       $txrx($protocol,auth_credentials) \
				       $txrx($protocol,auth_verifier)]]]
}

# 
# txrxInterpretReplyHeader --
#   Analyze a reply received from a server.
#
# Arguments:
#   Reply header as a Tcl data structure.
#   
# Result:
#   -code error if an error occured during the call 
#   or ok if call was succesfull.
#
# Side effects:
#   Updates xid and authorization values in global array txrx.

proc txrxInterpretReplyHeader {protocol reply} {
    global txrx
    set xid [lindex $reply 0]

    if {$xid != $txrx($protocol,xid)} {
        return -code error -errorinfo "error processing reply header" \
                "sequence number $xid does not match"
    }
    # Even if an error will occur later increase the sequence number.
    incr txrx($protocol,xid)

    set body [lindex $reply 1]
    if [string compare REPLY [lindex $body 0]] {
	return -code error -errorinfo "expected REPLY found [lindex $body 0]" \
                "RPC call header found, error while processing RPC"
    }

    set reply_body [lindex $body 1]
    set stat [lindex $reply_body 0]
    if [string compare $stat MSG_ACCEPTED] {
        set deny_body [lindex $reply_body 1]
        set reject_stat [lindex $deny_body 0]
        if {[string compare $reject_stat RPC_MISMATCH] == 0} {
           set lowhigh [lindex $deny_body 1]
           return -code error \
                  -errorinfo "message denied, incorrect RPC version $lowhigh" \
                "incorrect RPC version, error while processing RPC"
        }
        set auth_stat [lindex $deny_body 1]
 	return -code error \
               -errorinfo "message denied, authorization $auth_stat" \
                "rejected authorization $auth_stat, error while processing RPC"
    }

    set areply [lindex $reply_body 1]
    set txrx($protocol,reply_verifier) [lindex $areply 0]
    set astat [lindex [lindex $areply 1] 0]
    if [string compare $astat SUCCESS] {
	return -code error "error on the server [lindex $areply 1]"
    }

    return
}

# 
# txrxInterpretCallHeader --
#   Analyze an incoming RPC header and determine the procedure to call.
#
# Arguments:
#   Header as a Tcl data structure.
#   
# Result:
#   -code error if an error occured during the call 
#   or ok if call was succesfull.
#
# Side effects:
#   None.

proc txrxInterpretCallHeader {protocol header} {
    global txrx
    set msgBody [lindex $header 1]
    set msgType [lindex $msgBody 0]
    set callBody [lindex $msgBody 1]
    set rpcVersion [lindex $callBody 0]
    set prog [lindex $callBody 1]
    set vers [lindex $callBody 2]
    set proc [lindex $callBody 3]

    if [string match $msgType "REPLY"] {
        return -code error \
               -errorinfo "message header is a REPLY header -- $header" \
               [list REPLY \
                    [list MSG_ACCEPTED \
                         [list "AUTH_NONE {}" [list SYSTEM_ERR {}]]]]
    }
    if {$rpcVersion != 2} {
        return -code error \
               -errorinfo "incorrect RPC protocol version number -- $header" \
               [list REPLY \
                    [list MSG_DENIED \
                         [list RPC_MISMATCH [list 2 2]]]]
    }

    # Note: since the header was read and the RPC version is 2
    # this means at least a header MUST be sent in reply to the client.
    # Get procedure info.
    if [catch "$protocol info procedure [list [list $proc $vers $prog]]" procInfo] {
        return -code error \
               -errorinfo "procedure does not exist -- $header" \
               [list REPLY \
                    [list MSG_ACCEPTED \
                         [list "AUTH_NONE {}" \
                               [list PROG_MISMATCH [list 1 1]]]]]
    }
    if [catch "txrxCheckAuth [list [lindex $callBody 4]] [list [lindex $callBody 5]]" authInfo] {
        return -code error \
               -errorinfo "authentication failed -- $header" \
               [list REPLY \
                    [list MSG_DENIED \
                         [list AUTH_ERROR AUTH_FAILED]]]
    }
    #
    return $procInfo
}

#
# txrxCheckAuth --
#   Verify authorization for a client call.
#
# Parameters:
#   cred - authorization credential
#   verf - authorization verifier
#
# Result:
#   A standard Tcl result.

proc txrxCheckAuth {cred verf} {
    return
}       

#
# txrxSendErrorReply --
#   Reply to client by sending an error message encoded in the RPC header.
# 
# Parameters:
#   protocol - RPC communication protocol command
#   chan     - communication channel
#   header   - RPC header to send back to the client 
#
# Result:
#   None.
#
# Side effects:
#   Input buffer is cleared, data is written to the channel. If an error
#   occurs it is caught and ignored.

proc txrxSendErrorReply {protocol chan header} {
    catch "$protocol buffer reset
           $protocol buffer init write $chan
           $protocol write __rpc [list $header]
           $protocol buffer flush
    "
    flush $chan
    catch "$protocol buffer reset"
}

# 
# txrxReadCallArguments --
#   Read arguments from the network and convert to Tcl data.
#
# Parameters:
#   protocol - RPC protocol guiding the conversion
#   info     - information about the procedure 
#
# Result:
#   A list with the arguments.

proc txrxReadCallArguments {protocol procInfo} {
    set argTypes [lrange [lindex $procInfo 1] 1 end]
    set pArgs {}
    foreach argType $argTypes {
	if {$argType == 1} {
	    # VOID format arguments are not supposed to have a matching
	    # actual argument. Use the same actual argument for the
            # next formal argument.
        } else {
	    # Non-VOID argument. Must read some data from the network.
            if [catch "$protocol read $argType" argVal] {
                return -code error -errorinfo $argVal \
		  {REPLY {MSG_ACCEPTED { {AUTH_NONE {}} {GARBAGE_ARGS {}}}}}
            }
            lappend pArgs $argVal
        }
    }
    return $pArgs
}

# 
# txrxCall --
#   Call a remote procedure via the portmapper on the remote host.
#   The procedure inquires the portmapper on the remote host for
#   the transport address (port) where the server is registered.
#   It then opens a connection to the server using the specified
#   port and stores the connection as a private variable under
#   index "$protocol,channel" in array txrx. It the calls CallDirect
#   and returns the result. 
#
# Parameters:
#   protocol
#   hostname 
#   procedure - to call in usual format
#   args      - arguments for the call
#
# Result:
#   Result of remote call or an error
#
# Side effects:
#   Opens a communication channel to the server. This communication
#   channel is maintained until txrxDelete is called.

proc txrxCall {protocol host procedure args} {
    global txrx
    set procInfo [$protocol info procedure $procedure]
    set pvpInfo [lindex $procInfo 0]
    set program [lindex $pvpInfo 2]
    set version [lindex $pvpInfo 1]
 
    # Check if a communication channel is not open for the
    # program/version pair on the given host.

    if [catch "set txrx($protocol,channel,$host,$program,$version)" chan] {
	# No channel available for the host,program,version specified.
        set port [txrxGetServiceAddress $protocol $host $program $version 6]
        if {$port == -1} {
	    return -code error "error while contacting portmaper on host $host"
        }
        if {$port == 0} {
            return -code error \
		"no such program $program with version $version on host $host"
        }
        # Obtained a legal port number, try to open it.
        set chan [socket $host $port]
        # Making the channel non-blocking avoids any problems with
        # suddenly autistic servers. This should be a user configurable
        # option, as slow servers may send data much to slow and may
        # cause a data not available error. This error happens much more
        # often than the other case. Also, flushing socket buffers 
        # in non-blocking mode interferes with buffer management.
        # fconfigure $chan -blocking no

        # This option is mandatory.
        fconfigure $chan -translation binary
        set txrx($protocol,channel,$host,$port) $chan
    }
    eval txrxCallDirect $protocol $chan [list $pvpInfo] $args
}

# 
# txrxRegister --
#   Register a server with the local portmapper.
#
# Parameter:
#   protocol
#   channel    - must support getting the channel identification.
#   program    - number of program to register
#   version    - number of version to register
#
# Returns:
#   Nothing. An error occurs if portmapper cannot be contacted.

proc txrxRegister {protocol channel program version} {
    global txrx
    # Check if the program and version given exist in the protocol.
    set found 0
    foreach sym [$protocol info symbols] {
        if {[lindex $sym 0] == "PROGRAM"} {
            if {[lindex $sym 2] == $program} {
		# Found program number, look for the version.
                set versions [lrange $sym 3 end] 
                foreach vers $versions {
		    if {[lindex $vers 2] == $version} {
			set found 1 
			break
                    }
                }
		if {$found} {break}
            }
        }
    }
    if {$found == 0} {
	return -code error \
	 	"no such program $program, $version in protocol $protocol"
    }
    # 
    # Get port - if channel is not tcp (and hopefully udp) will generate
    # an error.
    set serverPort [lindex [fconfigure $channel -sockname] 2]

    # Open temporary channel to portmapper.
    if [catch "socket localhost 111" pmapChannel] {
        return -code error "this computer does not run the portmapper service"
    }
    fconfigure $pmapChannel -translation binary
    #
    # Create a special protocol for the portmapper.
    set pmapProtocol [txrxCreate "
	struct mapping { unsigned int prog; unsigned int vers; unsigned int prot; unsigned int port; };
        program P { version V {bool  SET(mapping) = 1; } = 2; } = 100000; "]
    # Perform a callDirect to register with the portmapper.
    if [catch "txrxCallDirect $pmapProtocol $pmapChannel SET \
	    [list [list $program $version 6 $serverPort]]" answer] {
        catch "txrxDelete $pmapProtocol" res
        close $pmapChannel
        return -code error "cannot register service with portmapper"
    } 
    catch "txrxDelete $pmapProtocol" res
    close $pmapChannel
    if {$answer == "FALSE"} {
        # Portmapper refused registration. Let the user know.
        return -code error \
           "portmap application refused \
to register $program, $version on port $serverPort"
    }
    # Success, program was registered.
    # Store this information (so that deregistration is possible on
    # server shutdown).
    set txrx($protocol,registered,$program,$version) $serverPort
}

# 
# txrxDeregister --
#   Remove portmapper information about a service. This is usually
#   performed when the server is shutdown. The portmapper has to be
#   informed about the event.
# Parameters:
#   protocol 
#   program -
#   version -  program and version number being deregistered.
# 
# Result:
#   A standard Tcl result.
#
# Note:
#   It is the responsibility of the service implementor to close the
#   server channel after deregistration.

proc txrxDeregister {protocol program version} {
    global txrx
    set port $txrx($protocol,registered,$program,$version)
    if [catch "socket localhost 111" chan] {
        return -code error "cannot contact portmapper service"
    }
    fconfigure $chan -translation binary
    # Create a protocol with a single procedure.
    set pmapProtocol [txrxCreate "
	struct mapping {int prog;int vers;int prot;int port;};
        program P { version V { bool UNSET(mapping) = 3;} = 2;} = 100000;"]
    
    # Perform a callDirect to the remote host.
    if [catch "txrxCallDirect $pmapProtocol $chan UNSET \
	    [list [list $program $version 6 $port]]" answer] {
	# Cleanup and return error message.
        txrxDelete $pmapProtocol
        close $chan
        return -code error "error while calling portmapper"
    }
    txrxDelete $pmapProtocol
    flush $chan
    close $chan
    unset txrx($protocol,registered,$program,$version)

    return $answer
}

#  
#
#
# txrxGetServiceAddress --
#   Contact portmapper service on socket 111 on remote host and obtain
#   the port number of the remote host where the service can be contacted.
#
# Parameters:
#   protocol - encapsulation of protocol description.
#   host     - name of remote computer.
#   program  - number of program.
#   version  - version of program.
#   tran     - transport code (6 for TCP, 17 for UDP).
#
# Return:
#   Address (port) on remote where the service resides.

proc txrxGetServiceAddress {protocol host program version tran} {
    set chan [socket $host 111]
    fconfigure $chan -translation binary
    # Create a protocol with a single procedure.
    set pmapProtocol [txrxCreate "
	struct mapping {
	    int prog;
	    int vers;
	    int prot;
	    int port;
        };
        program PMAP_PROG {
	    version PMAP_VERS {
		int        PMAPPROC_GETPORT(mapping)   = 3;
	    } = 2;
        } = 100000;
    "] 
    #
    # Perform a callDirect to the remote host.
    # What it should do is PMAPPROC_DUMP then search the list by itself.
    # Some old portmappers (v2) return 111 even when BIND (v3) is not
    # supported. It seems they only match the program number.
    # And when the actual call to service is sent, they just don't
    # reply - causing a timout to occur.
    if [catch "txrxCallDirect $pmapProtocol $chan PMAPPROC_GETPORT \
	    [list [list $program $version $tran 0]]" answer] {
        # Store error information for the user to read.
        global errorInfo
        set errorInfo $answer
        set answer -1
    }

    # Cleanup.
    catch "close $chan; txrxDelete $pmapProtocol"
    return $answer
};
    
