/* 
 * txrxConv.c --
 *
 *    Contains conversion code to/from XDR format and to/from TCL
 *    data structures.
 *
 * See the file "license.terms" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * SCCS: @(#) txrxConv.c 1.1 96/08/13 15:18:57
 */

#include "txrxInt.h"           /* Internal defintions. */

/*
 * Local functions.
 */
static int                  TypeToXDR _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                                   char *listPtr,
                                                   int addr));
static int                  TypeToList _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                                    int addr));

static int                  CheckAddress _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                                      int addr));

static char *               CharsToByte _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                                     char *listPtr,
                                                     int *n));

static char *               BytesToChar _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                                     char *bytePtr,
                                                     int n));

/*
 * Buffer management.
 */
static int                  DataRealloc _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                                     int incr));

static int                  DataPutInt _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                                    int n));
static int                  DataGetInt _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                                    int *n));
static int                  DataGet4 _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                                  char *buf));
static int                  ReadRecordHeader _ANSI_ARGS_((Txrx_Tape *tapePtr));
static int                  WriteRecordHeader _ANSI_ARGS_((Txrx_Tape *tapPtr));

static int                  DataPutDouble _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                                       double d));
static int                  DataGetDouble _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                                       double *d));
static int                  DataPutChars _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                                      int n,
                                                      char *bytePtr));
static int                  DataGetChars _ANSI_ARGS_((Txrx_Tape *tapePtr,
                                                      int n,
                                                      char *bytePtr));
/*
 * Data conversion functions.
 */
static void                 MakeInt _ANSI_ARGS_((Txrx_Tape *tapePtr, 
                                                 unsigned int n,
                                                 unsigned int *data));
static void                 MakeDouble _ANSI_ARGS_((Txrx_Tape *tapePtr, 
                                                    double d,
                                                    unsigned int *data));
/*
 * Local variables.
 */

/*
 *----------------------------------------------------------------------
 *
 * Txrx_Write --
 *
 *    Convert a Tcl data structure to network format and store it
 *    into the output buffer. The Tcl data structure is assumed to
 *    comply with its description stored on tape.
 *
 * Results:
 *    A standard Tcl result.
 *
 * Side effects:
 *    Data is sent to the data buffer in XDR format.
 *
 * Note:
 *    Data is stored in a buffer and then flushed to the channel
 *    in a single write operation - to accomodate stream oriented
 *    channels that use the record marking standard.
 *----------------------------------------------------------------------
 */

int
Txrx_Write(tapePtr, typeNamePtr, listPtr)
     Txrx_Tape *tapePtr;
     char *typeNamePtr;
     char *listPtr;
{
    SymbolData *symbolPtr;
    int addr;

    if (listPtr == NULL) {
        Tcl_AppendResult(tapePtr->interp, "data pointer is NULL",
                         (char *)NULL);
        return TCL_ERROR;
    }
    if (Tcl_GetInt(tapePtr->interp, typeNamePtr, &addr) == TCL_ERROR) {
        Tcl_ResetResult(tapePtr->interp);
        /*
         * Search the type name into the symbol table.
         */
        if ((symbolPtr = Txrx_GetSymbol(tapePtr, typeNamePtr)) == NULL) {
            return TCL_ERROR;
        }
        addr = symbolPtr->codeAddress;
    } else {
        if (CheckAddress(tapePtr, addr) == TCL_ERROR) {
            return TCL_ERROR;
        }
    }
    /*
     * Initialize data buffer.
     */
    if (tapePtr->dataOp != SEND) {
        Tcl_AppendResult(tapePtr->interp, "cannot write using input buffer",
                         (char *)NULL);
        return TCL_ERROR;
    }
    
    if (Txrx_ToXDR(tapePtr, listPtr, addr) == TCL_ERROR) {
        return TCL_ERROR;
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Txrx_Read --
 *
 *    Read data from a channel and convert it to a Tcl list.
 *
 * Results:
 *    A standard Tcl result.
 *
 * Side effects:
 *    interp->result contains the result of the read in human
 *    readable format.
 *
 *----------------------------------------------------------------------
 */

int                 
Txrx_Read(tapePtr, typeNamePtr)
     Txrx_Tape *tapePtr;
     char *typeNamePtr;
{
    SymbolData *symbolPtr;
    int addr;

    if (Tcl_GetInt(tapePtr->interp, typeNamePtr, &addr) == TCL_ERROR) {
        Tcl_ResetResult(tapePtr->interp);
        /*
         * Search the type name into the symbol table.
         */
        if ((symbolPtr = Txrx_GetSymbol(tapePtr, typeNamePtr)) == NULL) {
            return TCL_ERROR;
        }
        addr = symbolPtr->codeAddress;
    } else {
        if (CheckAddress(tapePtr, addr) == TCL_ERROR) {
            return TCL_ERROR;
        }
    }
    if (tapePtr->dataOp != RECEIVE) {
        Tcl_AppendResult(tapePtr->interp, "cannot read using ",
                         tapePtr->dataOp == SEND?"output":"uninitialized",
                         " buffer", (char *)NULL);
        return TCL_ERROR;
    }
    
    if (Txrx_ToList(tapePtr, addr) == TCL_ERROR) {
        return TCL_ERROR;
    }
    return TCL_OK;
}
/*
   Notes:
   XDR in the following means the channel and the fact that all data to/from 
   the channel is in network independent format.

   Conversion to XDR is straightforward - no buffers are allocated,
   the tape is used to grab data and send it to the network.

   Conversion from XDR is more involved. Besides managing a dynamic
   buffer, the program needs to store pointers (memory addresses) in the
   newly created data structure. As the C data structure grows there
   will be reallocations of the buffer. This means that all pointers will
   become automatically invalid. Fortunately, a patching procedure
   can be applied when the conversion is finished by storing the 
   location of all pointers into a table and by keeping in each pointer
   the relative displacement of the target location with respect to
   the beginning of the buffer. The patching adds to the relative
   offsets the starting address of the buffer.
   
   Another issue is how the caller will dispose of the allocated buffer
   when it is done. The solution is to perform a ckfree on the pointer 
   returned as result. A package provided function may be safer - worth
   considering as alternative.

   A week later:
   decided to use the exact opposite. Buffer is allocated
   to store the XDR data and send it in one call to the channel. When
   receiving, data is read in large chunks and converted as needed.
   Buffer management is hidden from the conversion layer.
   No more list to C conversions or arbitrary C to list - complicates
   things too much.
   Protocol tape keeps information about the buffer.
*/

/*
 *----------------------------------------------------------------------
 *
 * Txrx_ToXDR --
 *
 *    Convert a Tcl list to XDR format. The conversion
 *    is guided by the description of the data type stored in bytecode
 *    format.
 *
 * Results:
 *    A standard Tcl result.
 *
 * Side effects:
 *    None.
 *
 *----------------------------------------------------------------------
 */

int
Txrx_ToXDR(tapePtr, listPtr, addr)
     Txrx_Tape *tapePtr;
     char *listPtr;
     int addr;
{
    int n;
    unsigned int un;
    union { float f; unsigned int u;} mix;
    double d;
    char buf[32];
    CellType cellType;

    switch (addr) {
      case TIMPOSSIBLE:
        Tcl_AppendResult(tapePtr->interp, "invalid address on tape",
                         (char *)NULL);
        return TCL_ERROR;
      case TVOID:
        break;
      case TINT:
        if (Tcl_GetInt(tapePtr->interp, listPtr, &n) == TCL_ERROR) {
            return TCL_ERROR;
        }
        tapePtr->dataValue = n;
        DataPutInt(tapePtr, n);
        break;
      case TUINT:
        if (Tcl_GetInt(tapePtr->interp, listPtr, (int *)&un) == TCL_ERROR) {
            return TCL_ERROR;
        }
        tapePtr->dataValue = un;
        DataPutInt(tapePtr, un);
        break;
      case TBOOL:
        if (Tcl_GetBoolean(tapePtr->interp, listPtr, &n) == TCL_ERROR) {
            return TCL_ERROR;
        }
        tapePtr->dataValue = n;
        DataPutInt(tapePtr, n);
        break;
      case TFLOAT:
        if (Tcl_GetDouble(tapePtr->interp, listPtr, &d) == TCL_ERROR) {
            return TCL_ERROR;
        }
        mix.f = (float)d;
        DataPutInt(tapePtr, mix.u);
        break;
      case TDOUBLE:
        if (Tcl_GetDouble(tapePtr->interp, listPtr, &d) == TCL_ERROR) {
            return TCL_ERROR;
        }
        DataPutDouble(tapePtr, d);
        break;
      case TQUAD:
      case THYPER:
      case TUHYPER:
        Tcl_AppendResult(tapePtr->interp,
                         "sending hyper and quad data types ",
                         "is not implemented", (char *)NULL);
        return TCL_ERROR;
        break;
      default:
        if (addr < CODE_START || addr >= tapePtr->endCode) {
            Tcl_AppendResult(tapePtr->interp,
                             "invalid address in code", (char *)NULL);
            return TCL_ERROR;
        }
        if (tapePtr->codePtr[addr] != CELL_START) {
            Tcl_AppendResult(tapePtr->interp,
                             "no cell start marker", (char *)NULL);
            return TCL_ERROR;
        }
        /* 
         * Retrieve the type of a compiled symbol.
         */
        cellType = (CellType) tapePtr->codePtr[addr + 1];
        switch (cellType) {
          case TENUM:
          case TOPAQUE:
          case TSTRING:
          case TARRAY:
          case TSTRUCT:
          case TUNION:
          case TPOINTER:
            if (TypeToXDR(tapePtr, listPtr, addr)
                == TCL_ERROR) {
                return TCL_ERROR;
            }
            break;
          case TPROGRAM:
          case TVERSION:
          case TPROC:
          case TMODULE:
            Tcl_AppendResult(tapePtr->interp, 
                             "not implemented: proc, etc. call\n",
                             (char *)NULL);
          default:
            sprintf(buf, "%d", cellType);
            Tcl_AppendResult(tapePtr->interp, "cannot process cell type",
                             buf, (char *)NULL);
            return TCL_ERROR;
        }
        break;
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Txrx_ToList --
 *
 *    Build a Tcl data structure by reading it from the network.
 *
 * Results:
 *    A standard Tcl result.
 *
 * Side effects:
 *    Data is read from the channel, converted into a Tcl list
 *    and stored in interp->result. For a description of the
 *    format of the result read the txrx manual page.
 *
 *----------------------------------------------------------------------
 */

int
Txrx_ToList(tapePtr, addr)
     Txrx_Tape *tapePtr;
     int addr;
{
    CellType cellType;
    int n;
    float f;
    double d;
    char buf[TCL_DOUBLE_SPACE];
    
    switch (addr) {
      case TIMPOSSIBLE:
        Tcl_AppendResult(tapePtr->interp, "invalid address on tape",
                         (char *)NULL);
        return TCL_ERROR;
      case TVOID:
        break;
      case TINT:
        if (DataGetInt(tapePtr, &n) == TCL_ERROR) {
            return TCL_ERROR;
        }
        tapePtr->dataValue = n;
        sprintf(buf, "%d", n);
        Tcl_AppendElement(tapePtr->interp, buf);
        break;
      case TUINT:
        if (DataGetInt(tapePtr, &n) == TCL_ERROR) {
            return TCL_ERROR;
        }
        tapePtr->dataValue = (unsigned int)n;
        sprintf(buf, "%u", (unsigned int)n);
        Tcl_AppendElement(tapePtr->interp, buf);
        break;
      case TBOOL:
        if (DataGetInt(tapePtr, &n) == TCL_ERROR) {
            return TCL_ERROR;
        }
        tapePtr->dataValue = n;
        sprintf(buf, "%s", n == 0?"FALSE":"TRUE");
        Tcl_AppendElement(tapePtr->interp, buf);
        break;
      case TDOUBLE:
        if (DataGetDouble(tapePtr, &d) == TCL_ERROR) {
            return TCL_ERROR;
        }
        Tcl_PrintDouble(tapePtr->interp, d, buf);
        Tcl_AppendElement(tapePtr->interp, buf);
        break;
      case TFLOAT:
        if (DataGetInt(tapePtr, (int *)&f) == TCL_ERROR) {
            return TCL_ERROR;
        }
        Tcl_PrintDouble(tapePtr->interp, (double)f, buf);
        Tcl_AppendElement(tapePtr->interp, buf);
        break;
      case THYPER:
      case TUHYPER:
      case TQUAD:
        Tcl_AppendResult(tapePtr->interp,
                         "receiving hyper and quad data types ",
                         "is not implemented", (char *)NULL);
        return TCL_ERROR;
        break;
      default:
        if (addr < CODE_START || addr >= tapePtr->endCode) {
            Tcl_AppendResult(tapePtr->interp,
                             "invalid address in code", (char *)NULL);
            return TCL_ERROR;
        }
        if (tapePtr->codePtr[addr] != CELL_START) {
            Tcl_AppendResult(tapePtr->interp,
                             "no cell start marker", (char *)NULL);
            return TCL_ERROR;
        }
        /* 
         * Retrieve the type of a compiled symbol.
         */
        cellType = (CellType) tapePtr->codePtr[addr + 1];
        switch (cellType) {
          case TENUM:
          case TOPAQUE:
          case TSTRING:
          case TARRAY:
          case TSTRUCT:
          case TUNION:
          case TPOINTER:
            if (TypeToList(tapePtr, addr)
                == TCL_ERROR) {
                    return TCL_ERROR;
                }
            break;
            case TPROGRAM:
            case TVERSION:
            case TPROC:
            case TMODULE:
                Tcl_AppendResult(tapePtr->interp,
                                 "not implemented: proc, etc. call\n",
                                 (char *)NULL);
            default:
                sprintf(buf, "%d", cellType);
                Tcl_AppendResult(tapePtr->interp, "cannot process cell type",
                    buf, (char *)NULL);
                return TCL_ERROR;
        }
        break;
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * TypeToXDR --
 *
 *    Conversion from Tcl list to XDR data for a structured data type.
 *
 * Results:
 *    A standard Tcl result.
 *
 * Side effects:
 *    Data is written to the data buffer.
 *
 * Notes:
 * The representation chosen for XDR pointers in Tcl is
 * the Lisp like one - this in contrast to the way lists
 * are passed on the wire. An XDR NULL is a list of length
 * zero, while any list of length more than 1 means 
 * the pointer is not empty and the list itself is the value 
 * of the reference.
 * 
 * The alternative is to use TRUE/FALSE in a literal translation
 * of the standard (the RFC) but this is cumbersome. May be
 * added later as an alternative representation.
 *
 *----------------------------------------------------------------------
 */

static int
TypeToXDR(tapePtr, listPtr, addr)
     Txrx_Tape *tapePtr;
     char *listPtr;
     int addr;
{
    CellType cellType = (CellType) tapePtr->codePtr[addr + 1];
    Tcl_HashEntry *entryPtr;
    SymbolData *symbolPtr;
    int n;                       /* Number of fields, values, cases etc. */
    int argc;                    /* Number of elements in list.          */
    char **argv;                 /* Elements, making up the list.        */
    int i;
    int elmAddress;              /* Component byte code address.         */
    int len;                     /* Length of string (if given).         */
    int isVariable;              /* Byte data has variable length.       */
    static char buf[80];         /* Used to format messages and results. */
    char *bufPtr;                /* Intermediate buffer for opaques.     */
    int hasDefault;              /* Discriminated union has default.     */
    int discriminantAddress;     /* Location of case value.              */
    int defaultAddress;          /* Location where the default starts.   */
    int value;                   /* Store the discriminant value.        */
    char *caseListPtr;           /* Arguments for case or default 
                                  * alternative of a union. 
                                  */

    symbolPtr = (SymbolData *)tapePtr->codePtr[addr + 2];
    switch (cellType) {
      case TENUM:
        n = (int)tapePtr->codePtr[addr + 3];
        if (Tcl_GetInt(tapePtr->interp, listPtr, &n) == TCL_ERROR) {
            Tcl_ResetResult(tapePtr->interp);
            /* 
             * Not a numeric value for the enum, parse the string
             * and search the name among the names in the hash 
             * table.
             */
            if (Tcl_SplitList(tapePtr->interp, listPtr, &argc, &argv) ==
                TCL_ERROR) {
                ckfree((char *)argv);
                return TCL_ERROR;
            }
            if (argc > 1) {
                Tcl_AppendResult(tapePtr->interp, 
                                 "expected only one enumeration value",
                                 (char *)NULL);
                ckfree((char *)argv);
                return TCL_ERROR;
            }
            entryPtr = Tcl_FindHashEntry(symbolPtr->hashTablePtr, argv[0]);
            if (entryPtr == NULL) {
                Tcl_AppendResult(tapePtr->interp, 
                                 "\"", argv[0], 
                                 "\" is not an enumeration value",
                                 (char *)NULL);
                ckfree((char *)argv);
                return TCL_ERROR;
            }
            n = (int)Tcl_GetHashValue(entryPtr);
            ckfree((char *)argv);
        }
        tapePtr->dataValue = n;
        DataPutInt(tapePtr, n);
        break;
    case TSTRUCT:
        n = (int)tapePtr->codePtr[addr + 3];
        if (Tcl_SplitList(tapePtr->interp, listPtr, &argc, &argv)
            == TCL_ERROR) {
            return TCL_ERROR;
        }
        if (n != argc) {
            sprintf(buf, "expected %d fields, found %d", n, argc);
            Tcl_AppendResult(tapePtr->interp, buf, (char *)NULL);
            ckfree((char *)argv);
            return TCL_ERROR;
        }
        for (i = 0; i < n; i ++) {
            elmAddress = tapePtr->codePtr[addr + 4 + i];
            if (Txrx_ToXDR(tapePtr, argv[i], elmAddress) 
                == TCL_ERROR) {
                ckfree((char *)argv);
                return TCL_ERROR;
            }
        }
        ckfree((char *)argv);
        break;
    case TPOINTER:
        elmAddress = tapePtr->codePtr[addr + 3];
        if (Tcl_SplitList(tapePtr->interp, listPtr, &argc, &argv)
            == TCL_ERROR) {
            return TCL_ERROR;
        }
        if (argc == 0) {
            /*
             * The equivalent of a NULL pointer.
             */
            DataPutInt(tapePtr, 0);
        } else {
            DataPutInt(tapePtr, 1);
            if (Txrx_ToXDR(tapePtr, listPtr, elmAddress) == TCL_ERROR) {
                ckfree((char *)argv);
                return TCL_ERROR;
            }
        }
        ckfree((char *)argv);
        break;
    case TSTRING:
    case TOPAQUE:
        len = tapePtr->codePtr[addr + 3];
        isVariable = tapePtr->codePtr[addr + 4];
        if (cellType == TOPAQUE) {
            if ((bufPtr = CharsToByte(tapePtr, listPtr, &n)) == NULL) {
                return TCL_ERROR;
            }
        } else {
            n = strlen(listPtr);
            bufPtr = listPtr;
        }
        if (isVariable) {
            /*
             * Get the length of the string from the data structure.
             */
            if (len != 0) {
                /*
                 * There is an upper bound on the number of bytes to send.
                 */
                if (len > n) {
                    /*
                     * Input is less than maximum length, send only 'n' bytes.
                     */
                    len = n; 
                } else {
                    /* Input is more than maximum allowed length. The 
                     * standard required that at most 'len' bytes be sent.
                     */
                }
            } else {
                /*
                 * No limit on the numberof bytes to send.
                 */
                len = n;
            }
            DataPutInt(tapePtr, len);
        } else {
            if (cellType == TSTRING) {
                Tcl_AppendResult(tapePtr->interp,
                                 "strings cannot have fixed length",
                                 (char *)NULL);
                return TCL_ERROR;
            }
            if (n < len) {
                sprintf(buf, "%d bytes, expected %d", n, len);
                Tcl_AppendResult(tapePtr->interp, 
                                 "opaque has ", buf, (char *)NULL);
                ckfree((char *)bufPtr);
                return TCL_ERROR;
            }
        }
                
        /*
         * Add to data buffer the string.
         */
        DataPutChars(tapePtr, len, bufPtr);
        if (cellType == TOPAQUE) {
            /*
             * Free only for opaque - string does not allocate buffer.
             */
            ckfree((char *)bufPtr);
        }
        break;
      case TARRAY:
        len = tapePtr->codePtr[addr + 3];
        isVariable = tapePtr->codePtr[addr + 4];
        elmAddress = tapePtr->codePtr[addr + 5];
        
        if (Tcl_SplitList(tapePtr->interp, listPtr, &argc, &argv)
            == TCL_ERROR) {
            return TCL_ERROR;
        }
        if (isVariable) {
            if (len != 0) {
                if (len > argc) {
                    /*
                     * Less array elements in the list than maximum allowed.
                     * Send all the elements.
                     */
                    len = argc;
                } else {
                    /* 
                     * More array elements than expected. Send only the maximum
                     * allowed i.e. 'len'.
                     */
                }
            } else {
                len = argc;
            }
            DataPutInt(tapePtr, len);
        } else {
            if (argc < len) {
                ckfree((char *)argv);
                sprintf(buf, "expected %d elements in array, found only %d",
                        len, argc);
                Tcl_AppendResult(tapePtr->interp, buf, (char *)NULL);
                return TCL_ERROR;
            } else {
                /*
                 * Send only 'len' elements as required.
                 */
            }
        }
        if (argc < len) {
            Tcl_AppendResult(tapePtr->interp, "internal error sending array",
                             (char *)NULL);
            ckfree((char *)argv);
            return TCL_ERROR;
        }
        for (i = 0; i < len; i ++) {
            if (Txrx_ToXDR(tapePtr, argv[i], elmAddress) == TCL_ERROR) {
                ckfree((char *)argv);
                return TCL_ERROR;
            }
        }
        ckfree((char *)argv);
        break;
      case TUNION:
        discriminantAddress = tapePtr->codePtr[addr + 3];
        n = tapePtr->codePtr[addr + 4];
        hasDefault = tapePtr->codePtr[addr + 5];
        defaultAddress = tapePtr->codePtr[addr + 6];
        /*
         * Parse the input - it should be a list with at least one element,
         * the discriminant.
         */
        if (Tcl_SplitList(tapePtr->interp, listPtr, &argc, &argv)
            == TCL_ERROR) {
            return TCL_ERROR;
        }
        if (argc < 1) {
            Tcl_AppendResult(tapePtr->interp, "missing discriminant in union",
                             (char *)NULL);
            ckfree((char *)argv);
            return TCL_ERROR;
        }
        /*
         * Send the discriminant. Note that the value of the discriminant
         * is stored in tapePtr->dataValue.
         */
        if (Txrx_ToXDR(tapePtr, argv[0], discriminantAddress) == TCL_ERROR) {
            ckfree((char *)argv);
            return TCL_ERROR;
        }
        /*
         * Search in the code for the corresponding case value.
         */
        for (i = 7; i < (7 + 2 * n); i += 2) {
            value = tapePtr->codePtr[addr + i];
            elmAddress = tapePtr->codePtr[addr + i + 1];
            if (value == tapePtr->dataValue)  {
                break;
            }
        }
        if (i >= (7 + 2 * n)) {
            if (hasDefault == 0) {
                Tcl_AppendResult(tapePtr->interp, "case value \"",
                                 argv[0], 
                                 "\" invalid for union with no default",
                                 (char *)NULL);
                ckfree((char *)argv);
                return TCL_ERROR;
            }
            elmAddress = defaultAddress;
        }
        if (argc > 2) {
            /*
             * Assemble argv[1] through argv[argc - 1] into a list to
             * send.
             */
            caseListPtr = Tcl_Merge(argc - 1, argv + 1);
        } else {
            caseListPtr = argv[1] == NULL?"":argv[1];
        }
        if (Txrx_ToXDR(tapePtr, caseListPtr, elmAddress) == TCL_ERROR) {
            ckfree((char *)argv);
            if (argc > 2) {
                ckfree((char *)caseListPtr);
            }
            return TCL_ERROR;
        }
        ckfree((char *)argv);
        if (argc > 2) {
            ckfree((char *)caseListPtr);
        }
        break;
      default:
        Tcl_AppendResult(tapePtr->interp,
                         "C to XDR for unknown type not implemented", 
                         (char *)NULL);
        return TCL_ERROR;
        break;
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * TypeToList --
 *
 *    Convert XDR data into Tcl data.
 *
 * Results:
 *    A standard Tcl result.
 *
 * Side effects:
 *    interp->result is set to be the result of the conversion.
 *
 *----------------------------------------------------------------------
 */

static int
TypeToList(tapePtr, addr)
    Txrx_Tape *tapePtr;
    int addr;
{
    static char buf[32];
    CellType cellType = (CellType) tapePtr->codePtr[addr + 1];
    SymbolData *symbolPtr;       /* Pointer to symbol - needed when 
                                  * processing string values.
                                  */
    Tcl_HashSearch searchInfo;
    Tcl_HashEntry *entryPtr;
    int n;                       /* Number of fields, values, cases etc.   */
    char *valueName;             /* String with symbolic value.            */
    int argc;
    char **argv;                 /* Array storing intermediate results.    */
    int i, j;
    int elmAddress;              /* Location of field in bytecode.         */
    int len;                     /* Length of string (if given).           */
    int isVariable;              /* Byte data has variable length.         */
    char *bufferPtr;             /* String/opaque buffer.                  */
    char *bufPtr;                /* Opaque buffer - printable characters.  */
    int hasDefault;              /* Discriminated union has default.       */
    int discriminantAddress;     /* Location of case value.                */
    int defaultAddress;          /* Location where the default starts.     */
    char *valuePtr[2];           /* Discriminant and case arm storage.     */
    int value;                   /* Case value, numeric.                   */
    switch (cellType) {
      case TENUM:
        if (DataGetInt(tapePtr, &n) == TCL_ERROR) {
            return TCL_ERROR;
        }
        tapePtr->dataValue = n;
        sprintf(buf, "%d", n);
        valueName = buf;
        /*
         * Search the enum value in the enum table.
         */
        symbolPtr = (SymbolData *)tapePtr->codePtr[addr + 2];
        entryPtr = Tcl_FirstHashEntry(symbolPtr->hashTablePtr, &searchInfo);
        while (entryPtr) {
            if ((int) Tcl_GetHashValue(entryPtr) == n) {
                valueName = Tcl_GetHashKey(symbolPtr->hashTablePtr, entryPtr);
                break;
            }
            entryPtr = Tcl_NextHashEntry(&searchInfo);
        }
        Tcl_AppendElement(tapePtr->interp, valueName);
        break;
      case TSTRUCT:
        n = (int)tapePtr->codePtr[addr + 3];
        argc = n;
        argv = (char **)ckalloc(argc * sizeof(char *));
        for (i = 0; i < n; i ++) {
            elmAddress = tapePtr->codePtr[addr + 4 + i];
            if (Txrx_ToList(tapePtr, elmAddress) == TCL_ERROR) {
                for (j = 0; j < i; j ++) {
                    ckfree(argv[j]);
                }
                ckfree((char *)argv);
                return TCL_ERROR;
            }
            /*
             * interp->result contains a list for the subelement. Store
             * it in argv[i] and prepare the next iteration.
             */
            argv[i] = ckalloc(strlen(tapePtr->interp->result) + 1);
            strcpy(argv[i], tapePtr->interp->result);
            Tcl_ResetResult(tapePtr->interp);
        }
        /*
         * Prepare the composite (big) result.
         */
        Tcl_SetResult(tapePtr->interp, Tcl_Merge(argc, argv), TCL_DYNAMIC);
        for (i = 0; i < n; i ++) {
            ckfree(argv[i]);
        }
        ckfree((char *)argv);
        break;
      case TPOINTER:
        elmAddress = tapePtr->codePtr[addr + 3];
        if (DataGetInt(tapePtr, &n) == TCL_ERROR) {
            return TCL_ERROR;
        }
        if (n == 0) {
            Tcl_AppendResult(tapePtr->interp, "", (char *)NULL);
        } else {
            /*
             * Next location is the beginning of the result.
             */
            if (Txrx_ToList(tapePtr, elmAddress) == TCL_ERROR) {
                return TCL_ERROR;
            }
        }
        break;
      case TSTRING:
      case TOPAQUE:
        len = tapePtr->codePtr[addr + 3];
        isVariable = tapePtr->codePtr[addr + 4];
        if (isVariable) {
            if (DataGetInt(tapePtr, &n) == TCL_ERROR) {
                return TCL_ERROR;
            }
        } else {
            if (cellType == TSTRING) {
                Tcl_AppendResult(tapePtr->interp,
                                 "strings cannot have fixed length",
                                 (char *)NULL);
                return TCL_ERROR;
            }
            n = len;
        }
        /*
         * Allocate enough space to hold the read data padded up to
         * a multiple of sizeof(long) with NULLs.
         */
        bufferPtr = (char *)ckalloc((n + sizeof(long)) * sizeof(char));
        if (bufferPtr == NULL) {
            sprintf(buf, "%d", n);
            Tcl_AppendResult(tapePtr->interp, 
                             "cannot allocate string/opaque buffer of length ",
                             buf, (char *)NULL);
            return TCL_ERROR;
        }
        if (DataGetChars(tapePtr, n, bufferPtr) == TCL_ERROR) {
            ckfree((char *)bufferPtr);
            return TCL_ERROR;
        }
        bufferPtr[n] = '\0';
        if (cellType == TOPAQUE) {
            /* Strings are assumed to be ASCII strings - no chacks
             * are made on strings. Incoming strings with non-printable
             * characters will be sent to the user unprocessed.
             * Convert all non-printable characters to backslash sequences.
             * Note that the conversion is not the reverse of CharsToBytes.
             */
            if ((bufPtr = BytesToChar(tapePtr, bufferPtr, n)) == NULL) {
                ckfree((char *)bufferPtr);
                return TCL_ERROR;
            }
        } else {
            bufPtr = bufferPtr;
        }
        Tcl_AppendResult(tapePtr->interp, bufPtr, (char *)NULL);
        ckfree((char *)bufferPtr);
        if (cellType == TOPAQUE) {
            /*
             * Only opaques have the backslashes translated.
             */
            ckfree((char *)bufPtr);
        }
        break;
      case TARRAY:
        len = tapePtr->codePtr[addr + 3];
        isVariable = tapePtr->codePtr[addr + 4];
        elmAddress = tapePtr->codePtr[addr + 5];
        if (isVariable) {
            if (DataGetInt(tapePtr, &n) == TCL_ERROR) {
                return TCL_ERROR;
            }
        } else {
            n = len;
        }
        argc = n;
        argv = (char **)ckalloc(argc * sizeof(char *));
        for (i = 0; i < n; i ++) {
            if (Txrx_ToList(tapePtr, elmAddress) == TCL_ERROR) {
                for (j = 0; j < i; j ++) {
                    ckfree(argv[j]);
                }
                ckfree((char *)argv);
                return TCL_ERROR;
            }
            /*
             * interp->result contains a list for the subelement. Store
             * it in argv[i] and prepare the next iteration.
             */
            argv[i] = ckalloc(strlen(tapePtr->interp->result) + 1);
            strcpy(argv[i], tapePtr->interp->result);
            Tcl_ResetResult(tapePtr->interp);
        }
        /*
         * Prepare the composite (big) result.
         */
        Tcl_SetResult(tapePtr->interp, Tcl_Merge(argc, argv), TCL_DYNAMIC);
        for (i = 0; i < n; i ++) {
            ckfree(argv[i]);
        }
        ckfree((char *)argv);
        break;
      case TUNION:
        discriminantAddress = tapePtr->codePtr[addr + 3];
        n = tapePtr->codePtr[addr + 4];
        hasDefault = tapePtr->codePtr[addr + 5];
        defaultAddress = tapePtr->codePtr[addr + 6];

        /*
         * tapePtr->dataValue will be set to the numeric value of 
         * the discriminant, interp->result is set to a symbolic 
         * value.
         */
        if (Txrx_ToList(tapePtr, discriminantAddress) == TCL_ERROR) {
            return TCL_ERROR;
        }
        /*
         * Search the discriminant value among the case values.
         */
         for (i = 7; i < (7 + 2 * n); i += 2) {
            value = tapePtr->codePtr[addr + i];
            elmAddress = tapePtr->codePtr[addr + i + 1];
            if (value == tapePtr->dataValue)  {
                break;
            }
        }
        if (i >= (7 + 2 * n)) {
            if (hasDefault == 0) {
                /*
                 * Use the fact the interp->result already contains
                 * the symbolic representation of the value.
                 */
                Tcl_AppendResult(tapePtr->interp, " invalid as case value ",
                                 "for union with no default", (char *)NULL);
                return TCL_ERROR;
            }
            elmAddress = defaultAddress;
        }
        /*
         * Save the current interpreter result with the discriminant
         * value.
         */
        valuePtr[0] = ckalloc(strlen(tapePtr->interp->result) + 1);
        strcpy(valuePtr[0], tapePtr->interp->result);
        Tcl_ResetResult(tapePtr->interp);
        if (Txrx_ToList(tapePtr, elmAddress) == TCL_ERROR) {
            ckfree(valuePtr[0]);
            return TCL_ERROR;
        }
        valuePtr[1] = ckalloc(strlen(tapePtr->interp->result) + 1);
        strcpy(valuePtr[1], tapePtr->interp->result);
        Tcl_ResetResult(tapePtr->interp);
        Tcl_SetResult(tapePtr->interp, Tcl_Merge(2, valuePtr), TCL_DYNAMIC);

        /*
         * Cleanup.
         */
        ckfree(valuePtr[0]);
        ckfree(valuePtr[1]);
        break;
      default:
        Tcl_AppendResult(tapePtr->interp,
                         "C to XDR for unknown type not implemented", 
                         (char *)NULL);
        return TCL_ERROR;
        break;
    }
    return TCL_OK;
}

/*
 * Utility routines performing conversions between network and
 * machine format. Conversions are defined for elementary types.
 */

/*
 *----------------------------------------------------------------------
 *
 * MakeInt --
 *
 *    Convert a network/machine integer (big endian) into a
 *    machine/network integer.
 *
 * Results:
 *    None.
 *
 * Side effects:
 *    None.
 *
 *----------------------------------------------------------------------
 */

static void
MakeInt(tapePtr, n, data)
     Txrx_Tape *tapePtr;
     unsigned int n;
     unsigned int *data;
{
    if (tapePtr->endian == TXRX_BIGENDIAN) {
        *data = n;
    } else {
        *data = n << 24 | ((n << 8) & 0x00ff0000) | ((n >> 8) & 0x0000ff00) | n >> 24;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * MakeDouble --
 *
 *    Convert a double into an XDR array of two ints and viceversa.
 *
 * Results:
 *    None.
 *
 * Side effects:
 *    None.
 *
 *----------------------------------------------------------------------
 */

static void
MakeDouble(tapePtr, d, odata)
    Txrx_Tape *tapePtr;
    double d;
    unsigned int *odata;
{
    
    union {
        unsigned int idata[2];
        double data;
    } mix;
    
    if (tapePtr->endian == TXRX_BIGENDIAN) {
        mix.data = d;
        odata[0] = mix.idata[0];
        odata[1] = mix.idata[1];
    } else {
        mix.data = d;
        odata[1] = (mix.idata[0] << 24)
          | ((mix.idata[0] << 8) & 0x00ff0000)
          | ((mix.idata[0] >> 8) & 0x0000ff00)
          | (mix.idata[0] >> 24);
        
        odata[0] = (mix.idata[1] << 24)
          | (((mix.idata[1] << 8) & 0x00ff0000))
          | ((mix.idata[1] >> 8) & 0x0000ff00) 
          | (mix.idata[1] >> 24);
    }
}

/*
 *----------------------------------------------------------------------
 *
 * Txrx_GetSymbol --
 *
 *    Return a pointer to symbol corresponding to type name passed
 *    as argument. The protocol description where the type name 
 *    is assumed to be defined is stored in parameter tapePtr.
 *
 * Results:
 *    Pointer to symbol or NULL if symbol cannot be found.
 *
 * Side effects:
 *    None.
 *
 *----------------------------------------------------------------------
 */

SymbolData *
Txrx_GetSymbol(tapePtr, typeNamePtr)
    Txrx_Tape *tapePtr;
    char *typeNamePtr;
{
    Tcl_HashEntry *entryPtr;
    SymbolData *symbolPtr;

    if (typeNamePtr == NULL) {
        Tcl_AppendResult(tapePtr->interp, 
                         "type name given as argument is NULL",
                         (char *)NULL);
        return NULL;
    }   
    entryPtr = Tcl_FindHashEntry(&tapePtr->constTypeTable, typeNamePtr);
    if (   entryPtr == NULL 
        || ((symbolPtr = (SymbolData *)Tcl_GetHashValue(entryPtr)) == NULL)) {
        Tcl_AppendResult(tapePtr->interp, 
                         "type \"", typeNamePtr, "\" does not exist",
                         (char *)NULL);
        return NULL;
    }
    if (symbolPtr->type == SYM_CONST || symbolPtr->codeAddress == 0) {
        Tcl_AppendResult(tapePtr->interp,
            "type \"", typeNamePtr, "\" has no bytecode representation",
            (char *)NULL);
        return NULL;
    }
    return symbolPtr;
}

/*
 *----------------------------------------------------------------------
 *
 * CheckAddress --
 *
 *    Check if a bytecode address is valid (either predefined type or
 *    compiled).
 *
 * Results:
 *    A standard Tcl result.
 *
 * Side effects:
 *    None.
 *
 *----------------------------------------------------------------------
 */

static int
CheckAddress(tapePtr, addr)
     Txrx_Tape *tapePtr;
     int addr;
{
    switch (addr) {
      case TINT:
      case TUINT:
      case TFLOAT:
      case TBOOL:
      case TDOUBLE:
      case TVOID:
      case THYPER:
      case TUHYPER:
      case TQUAD:
        break;
      default:
        if (addr < CODE_START || addr >= tapePtr->endCode) {
            Tcl_AppendResult(tapePtr->interp, "invalid address in code",
                             (char *)NULL);
            return TCL_ERROR;
        }
        if (tapePtr->codePtr[addr] != CELL_START) {
            Tcl_AppendResult(tapePtr->interp,
                             "no cell start marker at address",
                             (char *)NULL);
            return TCL_ERROR;
        }
        break;
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * CharsToByte --
 *
 *    Convert a string containing backslash sequences into a 
 *    sequence of bytes where each backslash sequence has been
 *    converted into the corresponding byte value.
 *
 * Results:
 *    Pointer to byte buffer.
 *
 * Side effects:
 *    Allocates buffer with ckalloc.
 *
 *----------------------------------------------------------------------
 */

static char *
CharsToByte(tapePtr, listPtr, n)
     Txrx_Tape *tapePtr;
     char *listPtr;
     int *n;
{
    char *bufPtr = ckalloc(strlen(listPtr) + 1);
    char *outPtr;
    int m;

    if (bufPtr == NULL) {
        Tcl_AppendResult(tapePtr->interp, "cannot allocate buffer", 
                         (char *)NULL);
        return NULL;
    }
    for (outPtr = bufPtr; *listPtr; outPtr ++) {
        if (*listPtr == '\\') {
            *outPtr = Tcl_Backslash(listPtr, &m);
            listPtr += m;
        } else {
            *outPtr = *listPtr;
            listPtr ++;
        }
    }
    *outPtr = '\0';
    *n = (int)((long)outPtr - (long)bufPtr);
    return bufPtr;
}

/*
 *----------------------------------------------------------------------
 *
 * BytesToChar --
 *
 *    Convert a byte sequence in a backslash string.
 *
 * Results:
 *    Pointer to printable string.
 *
 * Side effects:
 *    Allocates string with ckalloc. 
 *
 *----------------------------------------------------------------------
 */

static char *
BytesToChar(tapePtr, bytePtr, n)
     Txrx_Tape *tapePtr;
     char *bytePtr;
     int n;
{
    char *bufPtr;
    char *outPtr;
    int i, count;
    /*
     * Compute an estimate of the number of characters needed.
     */
    for (i = 0, count = 0; i < n; i ++, count ++) {
        if (!isprint(bytePtr[i])) count += 3;
    }
    count ++;           /* Account for the terminating byte. */
    bufPtr = ckalloc(count * sizeof(char));
    if (bufPtr == NULL) {
        Tcl_AppendResult(tapePtr->interp, "cannot allocate buffer", 
                         (char *)NULL);
        return NULL;
    }
    for (outPtr = bufPtr, i = 0; i < n; outPtr ++, bytePtr ++, i ++) {
        if (!isprint(*bytePtr)) {
            sprintf(outPtr, "\\%03o", (int) (*bytePtr & 0xff));
            outPtr += 3;
        } else {
            *outPtr = *bytePtr;
        }
    }
    *outPtr = '\0';
    return bufPtr;
}

/*
 *----------------------------------------------------------------------
 *
 * Txrx_DataInit --
 *
 *    Initialize data buffer.
 *
 * Results:
 *    A standard Tcl result.
 *
 * Side effects:
 *    Modifies dataPtr in tapePtr to point to newly allocated memory
 *    area.
 *
 *----------------------------------------------------------------------
 */

int
Txrx_DataInit(tapePtr, chanPtr, sz)
     Txrx_Tape *tapePtr;
     Tcl_Channel chanPtr;
     int sz;
{
    if (tapePtr->dataOp != NONE) {
        Tcl_AppendResult(tapePtr->interp, 
                         "internal error - data buffer is already initialized",
                         " for ", tapePtr->dataOp==SEND?"write":"read",
                         (char *)NULL);
        /*printf("%s\n", tapePtr->interp->result);*/
        return TCL_ERROR;
    }
    tapePtr->dataOp = sz < 0?RECEIVE:SEND;
    tapePtr->chanPtr = chanPtr;
    switch (tapePtr->dataOp) {
      case RECEIVE:
        tapePtr->dataCount = -1;
        tapePtr->dataLast = 1;
        tapePtr->dataCount = 0;
        if (ReadRecordHeader(tapePtr) == TCL_ERROR) {
            return TCL_ERROR;
        }
        break;
      case SEND:
        /*
         * If estimate is zero, must allocate at least one storage unit.
         */
        sz ++;
        
        if (tapePtr->dataPtr != NULL) {
            Tcl_AppendResult(tapePtr->interp, 
                             "internal error - allocated data buffer", 
                             (char *)NULL);
            return TCL_ERROR;
        }
        /*
         * Allocate a new buffer of the requested size.
         */
        tapePtr->dataSize = 0;
        tapePtr->dataPtr = (long *)ckalloc(sz * sizeof(long));
        if (tapePtr->dataPtr == NULL) {
            Tcl_AppendResult(tapePtr->interp, "out of memory", (char *)NULL);
            return TCL_ERROR;
        }
        tapePtr->dataSize = sz;
        break;
      case NONE:
        Tcl_AppendResult(tapePtr->interp, "internal error - unexpected op",
                         (char *)NULL);
        return TCL_ERROR;
    }
    /*
     * Memory buffer is allocated, state variables have
     * been initialized.
     */
    tapePtr->dataPos = 0;
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Txrx_DataReset --
 *
 *    Free all memory allocated for the data buffer.
 *
 * Results:
 *    A standard Tcl result.
 *
 * Side effects:
 *    All memory in dataPtr of tapePtr is freed.
 *
 *----------------------------------------------------------------------
 */

int
Txrx_DataReset(tapePtr)
    Txrx_Tape *tapePtr;
{
    /*
     * Use flush to ensure no garbage is left to read or to write.
     */
    switch (tapePtr->dataOp) {
      case SEND:
        if (tapePtr->dataPtr == NULL) {
            Tcl_AppendResult(tapePtr->interp, 
                             "internal error - no data buffer to free", 
                             (char *)NULL);
            return TCL_ERROR;
        }
        ckfree((char *)tapePtr->dataPtr);
        break;
      case RECEIVE:
        /*
          printf("reset call, receive, pos %d, data count is %d %s\n",
               tapePtr->dataPos,
               tapePtr->dataCount, tapePtr->dataLast?"and last fragment":"");
         */
        if (Txrx_DataFlush(tapePtr) == TCL_ERROR) {
            return TCL_ERROR;
        }
        if (tapePtr->dataPtr != NULL) {
            Tcl_AppendResult(tapePtr->interp, 
                             "internal error - allocated data buffer to free", 
                             (char *)NULL);
            return TCL_ERROR;
        }
        break;
      default:
        Tcl_AppendResult(tapePtr->interp, 
                         "data buffer reset call on uninitialized buffer",
                         (char *)NULL);
        break;
    }
    tapePtr->dataPtr = NULL;
    tapePtr->dataSize = 0;
    tapePtr->dataPos = 0;
    tapePtr->dataOp = NONE;
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Txrx_DataFlush --
 *
 *    Read/Write data buffer to a channel.
 *
 * Results:
 *    A standard TCL result.
 *
 * Side effects:
 *    On write it writes to the channel all data in the 
 *    data buffer. On read it reads all the data avaliable
 *    from the channel. If channel is stream oriented it
 *    reads all the record fragments up to the last one.
 *
 *----------------------------------------------------------------------
 */

int
Txrx_DataFlush(tapePtr)
    Txrx_Tape *tapePtr;
{
    int len = tapePtr->dataPos * 4;  /* Length of data in bytes.             */
    int done;                        /* No more data to read.                */
    char *bufPtr;                    /* Data buffer used to read dummy data. */

    if (tapePtr->chanPtr == NULL) {
        Tcl_AppendResult(tapePtr->interp,
                         "internal error - no channel available", 
                         (char *)NULL);
        return TCL_ERROR;
    }
    
    if (tapePtr->dataOp == NONE) {
        Tcl_AppendResult(tapePtr->interp,
                         "internal error -- data buffer is not initialized",
                         ", no data to process", (char *)NULL);
        return TCL_ERROR;
    }
    
    switch (tapePtr->dataOp) {
      case SEND:
        if (WriteRecordHeader(tapePtr) == TCL_ERROR) {
            return TCL_ERROR;
        }
        if (Tcl_Write(tapePtr->chanPtr, (char *)tapePtr->dataPtr, len) 
            == TCL_ERROR) {
            Tcl_AppendResult(tapePtr->interp,
                             "error writing data to ouput channel",
                             (char *)NULL);
            return TCL_ERROR;
        }
        break;
      case RECEIVE:
        /*
         * Read fragments until the last fragment is read.
         */
        done = 0;
        while (!done) {
            /*
             * Some data is still available in this fragment.
             */
            if (tapePtr->dataCount > 0) {
                bufPtr = ckalloc(tapePtr->dataCount);
                if (Tcl_Read(tapePtr->chanPtr, bufPtr, tapePtr->dataCount)
                    != tapePtr->dataCount) {
                    Tcl_AppendResult(tapePtr->interp, 
                                     "cannot flush input record fragment",
                                     (char *)NULL);
                    ckfree(bufPtr);
                    return TCL_ERROR;
                }
                ckfree(bufPtr);
            }
            if (tapePtr->dataLast == 0) {
                /*
                 * The fragment just finished was not the last fragment.
                 * Read the next fragment header.
                 */
                if (ReadRecordHeader(tapePtr) == TCL_ERROR) {
                    return TCL_ERROR;
                }
            } else {
                done = 1;
            }
        }
        break;
      default:
        Tcl_AppendResult(tapePtr->interp, 
                         "internal error - data op on flush has invalid value",
                         (char *)NULL);
        break;
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * DataRealloc --
 *
 *    Reallocate data buffer. The data structure is
 *    stored as a component of a protocol tape.
 *
 * Results:
 *    A standard Tcl result.
 *
 * Side effects:
 *    Deallocates/reallocates memory.
 *
 *----------------------------------------------------------------------
 */

static int
DataRealloc(tapePtr, incr)
     Txrx_Tape *tapePtr;
     int incr;
{
    long *dataPtr;
    
    if (tapePtr->dataPtr == NULL) {
        /*
         * Most probably a programmer error - forgot to
         * perform a data init or no memory was 
         * available - try it again.
         */
        /*printf("realloc called with no data buffer allocated\n");*/
        return Txrx_DataInit(tapePtr, 
                             tapePtr->chanPtr, tapePtr->dataSize + incr);
    }
    dataPtr = (long *)ckrealloc((char *)tapePtr->dataPtr,
                                (tapePtr->dataSize + incr)
                                * sizeof(long));
    if (dataPtr == NULL) {
        Tcl_AppendResult(tapePtr->interp, "cannot reallocate data",
                         (char *)NULL);
        return 0;
    }
    tapePtr->dataSize += incr;
    tapePtr->dataPtr = dataPtr;
    return 1;
}

/*
 *----------------------------------------------------------------------
 *
 * WriteRecordHeader --
 *
 *    Write a record header and update the protocol description.
 *
 * Results:
 *    A standard Tcl result.
 *
 * Side effects:
 *    If the channel is stream oriented the length of the data 
 *    buffer in bytes is written to the channel in standard network
 *    order with high order bit set to 1 (last fragment).
 *
 *----------------------------------------------------------------------
 */

static int
WriteRecordHeader(tapePtr)
     Txrx_Tape *tapePtr;
{        
    unsigned int ibuf = tapePtr->dataPos * 4;
    Tcl_ChannelType *chanTypePtr = Tcl_GetChannelType(tapePtr->chanPtr);
    
    if (strcmp(chanTypePtr->typeName, "tcp") == 0) {
        /*
         * Make this fragment the only fragment.
         */
        ibuf |= 0x80000000;
        MakeInt(tapePtr, ibuf, &ibuf);
        if (Tcl_Write(tapePtr->chanPtr, (char *)&ibuf, 4) != 4) {
            Tcl_AppendResult(tapePtr->interp, 
                             "cannot write record header to output channel",
                             (char *)NULL);
            return TCL_ERROR;
        }
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * ReadRecordHeader --
 *
 *    Read a record header and update the protocol description.
 *
 * Results:
 *    A standard Tcl result.
 *
 * Side effects:
 *    None.
 *
 *----------------------------------------------------------------------
 */

static int
ReadRecordHeader(tapePtr)
     Txrx_Tape *tapePtr;
{        
    Tcl_ChannelType *chanTypePtr;
    unsigned int ibuf;
    
    chanTypePtr = Tcl_GetChannelType(tapePtr->chanPtr);
    /*
     * If the channel is stream oriented then it has to
     * have available as input the first fragment of data.
     */
    if (strcmp(chanTypePtr->typeName, "tcp") == 0) {
        if (tapePtr->dataCount != 0) {
            Tcl_AppendResult(tapePtr->interp, 
                             "internal error - data count is not zero",
                             (char *)NULL);
            /*printf("problem - dataCount is %d, pos %d\n", 
                   tapePtr->dataCount, tapePtr->dataPos);
                   */
            return TCL_ERROR;
        }
        tapePtr->dataCount = 0;
        /*
         * The record marking standard imposes that the 
         * first 4 bytes are the length of the incoming 
         * fragment.
         */
        if (Tcl_Read(tapePtr->chanPtr, (char *)&ibuf, 4) != 4) {
            Tcl_AppendResult(tapePtr->interp, 
                             "error reading marker from stream channel",
                             (char *)NULL);
            /*printf("error reading marker from stream channel\n");*/
            return TCL_ERROR;
        }
        MakeInt(tapePtr, ibuf, &ibuf);
        if (ibuf & 0x80000000) {
            tapePtr->dataLast = 1;
        } else {
            tapePtr->dataLast = 0;
        }
        ibuf &= 0x7fffffff;
        tapePtr->dataCount = ibuf;
        /*
          printf("%s record fragment read, data length %d\n", 
               tapePtr->dataLast?"last":"", tapePtr->dataCount);
               */
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * DataGet4 --
 *
 *    Read 4 bytes from the network, taking care of the the 
 *    record marking standard.
 *
 * Results:
 *    A standard Tcl result.
 *
 * Side effects:
 *    None.
 *
 *----------------------------------------------------------------------
 */

static int
DataGet4(tapePtr, buf)
     Txrx_Tape *tapePtr;
     char *buf;
{
    switch (tapePtr->dataOp) {
      case RECEIVE:
        if (tapePtr->dataCount <= 0) {
            /*
             * Read the header of the next fragment and update dataCount.
             */
            if (ReadRecordHeader(tapePtr) == TCL_ERROR) {
                return TCL_ERROR;
            }
        }
        /* 
         * Data count is not the number of bytes that will be read.
         */
        tapePtr->dataCount -= 4;
        if (Tcl_Read(tapePtr->chanPtr, buf, 4) != 4) {
            Tcl_AppendResult(tapePtr->interp, 
                             "error reading a four byte unit from channel",
                             (char *)NULL);
            return TCL_ERROR;
        }
        break;
      case SEND:
        Tcl_AppendResult(tapePtr->interp, "read operation while sending data",
                         (char *)NULL);
        break;
      default:
        Tcl_AppendResult(tapePtr->interp, 
                         "read operation with uninitialized data buffer",
                         (char *)NULL);
        break;
    }

    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * DataPutInt --
 *
 *    Add an integer to the data area.
 *
 * Results:
 *    A standard Tcl result.
 *
 * Side effects:
 *    Integer is written to the data area.
 *
 *----------------------------------------------------------------------
 */

static int
DataPutInt(tapePtr, n)
    Txrx_Tape *tapePtr;
    int n;
{
    long m;

    if (tapePtr->dataPos >= tapePtr->dataSize) {
        if (DataRealloc(tapePtr, DATA_INCREMENT) == TCL_ERROR) 
          return TCL_ERROR;
    }
    MakeInt(tapePtr, (unsigned int)n, (unsigned int *)&m);
    tapePtr->dataPtr[tapePtr->dataPos ++] = m;
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * DataGetInt --
 *
 *    Read an integer from the network.
 *
 * Results:
 *    A standard Tcl result.
 *
 * Side effects:
 *    None.
 *
 *----------------------------------------------------------------------
 */

static int
DataGetInt(tapePtr, n)
     Txrx_Tape *tapePtr;
     int *n;
{
    union {
        char buf[4];
        unsigned int n;
    } mix;

    if (DataGet4(tapePtr, mix.buf) == TCL_ERROR) {
        return TCL_ERROR;
    }
    MakeInt(tapePtr, mix.n, (unsigned int *)n);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * DataPutDouble --
 *
 *    Add a double to the data area.
 *
 * Results:
 *    A standard Tcl result.
 *
 * Side effects:
 *    Double is written to the data area.
 *
 *----------------------------------------------------------------------
 */

static int
DataPutDouble(tapePtr, d)
    Txrx_Tape *tapePtr;
    double d;
{
    unsigned int idata[2];

    if (tapePtr->dataPos + 1>= tapePtr->dataSize) {
        if (DataRealloc(tapePtr, DATA_INCREMENT) == TCL_ERROR) 
          return TCL_ERROR;
    }

    MakeDouble(tapePtr, d, idata);
    tapePtr->dataPtr[tapePtr->dataPos ++] = idata[0];
    tapePtr->dataPtr[tapePtr->dataPos ++] = idata[1];
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * DataGetDouble --
 *
 *    Read a double from the network.
 *
 * Results:
 *    A standard Tcl result.
 *
 * Side effects:
 *    None.
 *
 *----------------------------------------------------------------------
 */

static int
DataGetDouble(tapePtr, d)
     Txrx_Tape *tapePtr;
     double *d;
{
    union {
        char b[sizeof(double)];
        unsigned int idata[2];
        double d;
    } mix;
    

    if (   DataGet4(tapePtr, mix.b) == TCL_ERROR 
        || DataGet4(tapePtr, mix.b + 4) == TCL_ERROR) {
        return TCL_ERROR;
    }
    MakeDouble(tapePtr, mix.d, mix.idata);
    *d = mix.d;
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * DataPutChars --
 *
 *    Add a sequence of bytes to the data area and pad to multiple of 4.
 *
 * Results:
 *    A standard Tcl result.
 *
 * Side effects:
 *    None.
 *
 *----------------------------------------------------------------------
 */

static int
DataPutChars(tapePtr, n, bytePtr)
    Txrx_Tape *tapePtr;
    int n;
    char *bytePtr;
{
    int m;               /* Number of entries needed to store the string. */
    char *charPtr;       /* Pointer to data buffer as character buffer.   */
    int i, j;
    
    if (n == 0) {
        /*
         * Nothing to do.
         */
        return TCL_OK;
    }
    m = (n - 1) / sizeof(long) + 1;

    if (tapePtr->dataPos + m - 1 >= tapePtr->dataSize) {
        if (DataRealloc(tapePtr, 
                        (m < DATA_INCREMENT)?DATA_INCREMENT:m + 10) 
            == TCL_ERROR) 
            return TCL_ERROR;
    }
    charPtr = (char *)&tapePtr->dataPtr[tapePtr->dataPos];
    /*
     * Copy data byte by byte (copy in other units is not safe).
     */
    for (i = 0; i < n; i ++) {
        charPtr[i] = bytePtr[i];
    }
    /*
     * Pad up to multiple of sizeof(long).
     */
    for (j = 0; (n + j) % sizeof(long) != 0; j ++) {
        charPtr[n + j] = '\0';
    }
    tapePtr->dataPos += m;
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * DataGetChars --
 *
 *    Read a number of bytes from the network.
 *
 * Results:
 *    A standard Tcl result.
 *
 * Side effects:
 *    Always reads a multiple of sizeof(long) bytes. The caller
 *    must be careful to allocate enough storage.
 *
 *----------------------------------------------------------------------
 */

static int
DataGetChars(tapePtr, n, bytePtr)
     Txrx_Tape *tapePtr;
     int n;
     char *bytePtr;
{
    int m;               /* Number of entries needed to store the string. */
    int i;

    if (n == 0) {
        /*
         * Nothing to do.
         */
        return TCL_OK;
    }
    m = (n - 1) / sizeof(long) + 1;
    for (i = 0; i < m; i ++) {
        if (DataGet4(tapePtr, bytePtr + 4 * i) == TCL_ERROR) {
            return TCL_ERROR;
        }
    }
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Txrx_InfoProc --
 *
 *    Determine information about a given procedure: argument list
 *    proper name, etc.
 *
 * Results:
 *    A standard Tcl result.
 *
 * Side effects:
 *    interp->result will contain a string with the information as 
 *    a list with two sublists. The first sublist is 
 *    "procedure version program" as numeric values. The second 
 *    sublist is "result parameter1 ... parametern" where each
 *    parameteri is the bytecode address where it resides.
 *
 *----------------------------------------------------------------------
 */

int
Txrx_InfoProc(tapePtr, addr)
     Txrx_Tape *tapePtr;
     int addr;
{
    SymbolData *symbolPtr = (SymbolData *)tapePtr->codePtr[addr + 2];
    int i, argc = tapePtr->codePtr[addr + 3];
    char buf[128];

    sprintf(buf, "{%ld %ld %ld} {", symbolPtr->constValue, 
            symbolPtr->referenceValue->constValue,
            symbolPtr->referenceValue->referenceValue->constValue);
    Tcl_AppendResult(tapePtr->interp, buf, (char *)NULL);
    for (i = 0; i <= argc; i ++) {
        sprintf(buf, "%ld", tapePtr->codePtr[addr + 5 + i]);
        Tcl_AppendElement(tapePtr->interp, buf);
    }
    Tcl_AppendResult(tapePtr->interp, "}", (char *)NULL);
    return TCL_OK;
}

/*
 *----------------------------------------------------------------------
 *
 * Txrx_FindProc --
 *
 *    Determine if a given procedure exists.
 *
 * Results:
 *    A standard Tcl result.
 *
 * Side effects:
 *    Addr will contain the byte code address where the procedure
 *    description starts.
 *
 *----------------------------------------------------------------------
 */

int
Txrx_FindProc(tapePtr, namePtr, addr)
     Txrx_Tape *tapePtr;
     char *namePtr;
     int *addr;
{
    int argc;
    char **argv;
    int p, v, g;                   /* Numbers for proc, version, program.   */
    char *pPtr, *vPtr, *gPtr;      /* Symbolic names for the above.         */
    char buf[32];
    Tcl_HashEntry *entryPtr;
    Tcl_HashSearch searchInfo;
    SymbolData *symbolPtr;
    SymbolData *vsymbolPtr;
    SymbolData *psymbolPtr;
    
    /*
     * Parse the name.
     */
    if (Tcl_SplitList(tapePtr->interp, namePtr, &argc, &argv) == TCL_ERROR) {
        return TCL_ERROR;
    }
    p = v = g = -1;
    pPtr = vPtr = gPtr = "";
    
    if (argc != 3 && argc != 1) {
        Tcl_AppendResult(tapePtr->interp, "expected a list of the form \"",
                         "procedure version program\" or a procedure name", 
                         (char *)NULL);
        goto error;
    }
    if (argc == 1) {
        /*
         * Search the symbol table for the constant with that name.
         */
        entryPtr = Tcl_FindHashEntry(&tapePtr->constTypeTable, argv[0]);
        if (entryPtr == NULL) {
            Tcl_AppendResult(tapePtr->interp, "no such procedure name \"",
                             argv[0], "\"", (char *)NULL);
            goto error;
        }
        symbolPtr = (SymbolData *)Tcl_GetHashValue(entryPtr);
        if (   symbolPtr->type == SYM_CONST 
            && symbolPtr->referenceValue != NULL
            && symbolPtr->referenceValue->type == SYM_PROCEDURE) {
            *addr = symbolPtr->referenceValue->codeAddress;
            goto done;
        }
        Tcl_AppendResult(tapePtr->interp, "\"", argv[0], 
                         "\" is not a procedure name", (char *)NULL);
        goto error;
    }
    if (argc >= 1) {
        if (Tcl_GetInt(tapePtr->interp, argv[0], &p) == TCL_ERROR) {
            /*
             * Symbolic name - deal with it later.
             */
            Tcl_ResetResult(tapePtr->interp);
            pPtr = argv[0];
        }
    }
    if (argc >= 2) {
        if (Tcl_GetInt(tapePtr->interp, argv[1], &v) == TCL_ERROR) {
            /*
             * Symbolic name for version - deal with it later.
             */
            Tcl_ResetResult(tapePtr->interp);
            vPtr = argv[1];
        }
    }
    if (argc >= 3) {
        if (Tcl_GetInt(tapePtr->interp, argv[2], &g) == TCL_ERROR) {
            /*
             * Symbolic name for program - deal with it later.
             */
            
            Tcl_ResetResult(tapePtr->interp);
            gPtr = argv[2];
        }
    }

    /*
     * Start searching the symbol table for the program name.
     */
    entryPtr = Tcl_FirstHashEntry(&tapePtr->constTypeTable, &searchInfo);
    while (entryPtr) {
        symbolPtr = (SymbolData *)Tcl_GetHashValue(entryPtr);
        if (symbolPtr->type == SYM_PROGRAM) {
            if (   (g != -1 && g == symbolPtr->constValue)
                || (   g == -1 
                    && strcmp(Tcl_GetHashKey(&tapePtr->constTypeTable, 
                                             entryPtr),
                              gPtr) == 0)) {
                break;
            }
        }
        entryPtr = Tcl_NextHashEntry(&searchInfo);
    }
    if (entryPtr == NULL) {
        /*
         * Names do not match.
         */
        sprintf(buf, "%d", g);
        Tcl_AppendResult(tapePtr->interp, "program name \"",
                         g == -1?gPtr:buf,
                         "\" does not exist", (char *)NULL);
        goto error;
    }

    symbolPtr = (SymbolData *)Tcl_GetHashValue(entryPtr);
    /*
     * Search version name among the versions for the program.
     */
    entryPtr = Tcl_FirstHashEntry(symbolPtr->hashTablePtr, &searchInfo);
    while (entryPtr) {
        vsymbolPtr = (SymbolData *)Tcl_GetHashValue(entryPtr);
        if (vsymbolPtr->type != SYM_VERSION) {
            Tcl_AppendResult(tapePtr->interp, "internal error program symbol",
                             (char *)NULL);
            goto error;
        }
        if (   (v != -1 && v == vsymbolPtr->constValue)
            || (   v == -1 
                && strcmp(Tcl_GetHashKey(symbolPtr->hashTablePtr, entryPtr),
                          vPtr) == 0)) {
            break;
        }
        entryPtr = Tcl_NextHashEntry(&searchInfo);
    }
    if (entryPtr == NULL) {
        sprintf(buf, "%d", v);
        Tcl_AppendResult(tapePtr->interp, "version name \"",
                         v == -1?vPtr:buf,
                         "\" does not exist", (char *)NULL);
        goto error;
    }
    vsymbolPtr = (SymbolData *)Tcl_GetHashValue(entryPtr);
    /*
     * Search procedure symbol.
     */
    entryPtr = Tcl_FirstHashEntry(vsymbolPtr->hashTablePtr, &searchInfo);
    while (entryPtr) {
        psymbolPtr = (SymbolData *)Tcl_GetHashValue(entryPtr);
        if (psymbolPtr->type != SYM_PROCEDURE) {
            Tcl_AppendResult(tapePtr->interp, "internal error version symbol",
                             (char *)NULL);
            goto error;
        }
        if (   (p != -1 && p == psymbolPtr->constValue)
            || (   p == -1 
                && strcmp(Tcl_GetHashKey(symbolPtr->hashTablePtr, entryPtr),
                          pPtr) == 0)) {
            break;
        }
        entryPtr = Tcl_NextHashEntry(&searchInfo);
    }
    if (entryPtr == NULL) {
        sprintf(buf, "%d", v);
        Tcl_AppendResult(tapePtr->interp, "procedure name \"",
                         p == -1?pPtr:buf,
                         "\" does not exist", (char *)NULL);
        goto error;
    }
    psymbolPtr = (SymbolData *)Tcl_GetHashValue(entryPtr); 
    *addr = psymbolPtr->codeAddress;
  done:
    ckfree((char *)argv);
    return TCL_OK;
  error:
    ckfree((char *)argv);
    return TCL_ERROR;
}

/*
 *----------------------------------------------------------------------
 *
 * <procName> --
 *
 *    <description>
 *
 * Results:
 *    None.
 *
 * Side effects:
 *    None.
 *
 *----------------------------------------------------------------------
 */

static void
Proc()
{
}

